package SecureProtocol.BasicPrimitives;

import java.math.BigInteger;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.LinkedList;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.spec.SecretKeySpec;


/**
 * OT<sup>N</sup><sub>1</sub> protocol.
 * @author Ayman Jarrous
 * @version 1.0
 */

public class OT_N_1 {
	
	public final static int AESSise=128;
	private final static int FSize=20;
	protected final static SecureRandom RAND_GENERATOR=new SecureRandom();
	
	/**
	 * Random oracle simulator.
	 * @param key AES key
	 * @param x parameter to encrypt
	 * @param length size of the encryption.
	 * @return Encryption that depends on x and key of size length.
	 */
	public final static 
	BigInteger F(final BigInteger key, final BigInteger x, int length) {
		if(key==null) {
			System.err.println("(1.) OT_N_1, F: Null key");
			System.exit(-1);
		}
		if(x==null) {
			System.err.println("(2.) OT_N_1, F: Null parameter");
			System.exit(-1);
		}
		if(length<=0){
			System.err.println("(3.) OT_N_1, F: Invalid length");
			System.exit(-1);
		}
		
		BigInteger tempX=x;
		String strRes="";
		try {
			byte[] raw = key.toByteArray();
			SecretKeySpec skeySpec = new SecretKeySpec(raw, "AES");
			Cipher cipher = Cipher.getInstance("AES");
			cipher.init(Cipher.ENCRYPT_MODE, skeySpec);
			
			long counter=1;
			while(strRes.length()*Byte.SIZE <length) {
				tempX=tempX.setBit(8).add(BigInteger.valueOf(counter++));
				byte[] encrypted=cipher.doFinal(tempX.toString().getBytes());
				strRes=strRes+new String(encrypted);
			}
	    }
	    catch(InvalidKeyException ike) {
	    	System.err.print("(4.) OT_N_1, F: "+ike.getMessage());
	    	System.exit(-1);
	    }
	    catch(NoSuchPaddingException nspe) {
	    	System.err.print("(5.) OT_N_1, F: "+nspe.getMessage());
	    	System.exit(-1);
	    }
	    catch(BadPaddingException bpe) {
	    	System.err.print("(6.) OT_N_1, F: "+bpe.getMessage());
	    	System.exit(-1);
	    }
	    catch(NoSuchAlgorithmException nsae) {
	    	System.err.print("(7.) OT_N_1, F: "+nsae.getMessage());
	    	System.exit(-1);
	    }
	    catch(IllegalBlockSizeException ibse) {
	    	System.err.print("(8.) OT_N_1, F: "+ibse.getMessage());
	    	System.exit(-1);
	    }
		
	    return new BigInteger(strRes.getBytes());
	}

	//---------------------------------------------------------------
	
	/**
	 * XOR operation between two {@link BigInteger}.
	 * @param x first parameter
	 * @param y second parameter
	 * @return The result of xorring x and y
	 */
	public final static BigInteger 
	xorBigIntegers(final BigInteger x, final BigInteger y) {
		
		if(x==null || y==null) {
			System.err.println("OT_N_1, xorBigIntegers: One of the input is null");
			System.exit(-1);
		}
		
        byte inb1[]=x.toByteArray();
        byte inb2[]=y.toByteArray();
       
        byte[] result=null;
        int len=0;

        if(inb1.length>inb2.length) {
        	result=new byte[inb1.length];
        	len=inb2.length;
        } else {
        	result=new byte[inb2.length];
        	len=inb1.length;
        }
        
        for(int i=0; i<len; i++) {
        	result[result.length-1-i]=(byte)(inb1[inb1.length-1-i]^inb2[inb2.length-1-i]);
        }
        
        if(inb1.length > inb2.length) {
        	for(int i=0; i<inb1.length-len; i++) {        		
        		result[i]=inb1[i];
        	}
        } else {
        	for(int i=0; i<inb2.length-len; i++) {
        		result[i]=inb2[i];
        	}
        }
        return (new BigInteger(result));
    }
	
    //---------------------------------------------------------------
	
	
	/**
	 * Hashing the keys based on their locations
	 * @param keys list of keys to hash
	 * @return A list of hashed keys, every object on the list contains an array of hashed keys based on the location.
	 * 
	 */
	public final static LinkedList<BigInteger>
	prepareFKey(final LinkedList<BigInteger[]> keys) {
		
		if(keys==null) {
			System.err.println("OT_N_1, prepareFKey: Null keys list");
			System.exit(-1);
		}
		
		if(keys.isEmpty()) {
			return null;
		}
		int N=(int)Math.pow(2, keys.size());
		
		LinkedList<BigInteger> resList=new LinkedList<BigInteger>();
		int index[]=new int[keys.size()];
		BigInteger res=null;
		for(int i=0; i<N; i++) {
			for(int j=0; j<index.length; j++){
				index[j]=0;
			}
			
			String bin=Integer.toBinaryString(i);
			
			for(int k=0, j=bin.length()-1; j>=0; j--,k++) {
				if(bin.charAt(j)=='1') {
					index[k]=1;
				}
			}
			
			res=F(keys.get(0)[index[0]], BigInteger.valueOf(i),FSize);
			
			for(int j=1; j<keys.size(); j++) {
				res=xorBigIntegers(res, F(keys.get(j)[index[j]], BigInteger.valueOf(i),FSize));
			}
			resList.add(res);
		}
		return resList;
	}
	
	/**
	 * Encrypts a list of values using a list of hashed keys
	 * @param list list of items to encrypts
	 * @param keys hashed keys
	 * @return A list of encrypted items
	 */
	public final static LinkedList<BigInteger> 
	calculateYBasedonPrepareedKeys(LinkedList<BigInteger> list, final LinkedList<BigInteger> keys) {
		
		if(list==null) {
			System.err.println("(1.) OT_N_1, calculateYBasedonPrepareedKeys: Null list");
			System.exit(-1);
		}
		if(keys==null) {
			System.err.println("(2.) OT_N_1, calculateYBasedonPrepareedKeys: Null key list");
			System.exit(-1);
		}
		if(list.size()!=keys.size()) {
			System.err.println("(3.) OT_N_1, calculateYBasedonPrepareedKeys: List size is not equal to key list size");
			System.exit(-1);
		}
		
		
		LinkedList<BigInteger> res=new LinkedList<BigInteger>();
		for(int j=0; j<keys.size(); j++) {
			res.add(xorBigIntegers(list.get(j), keys.get(j)));
		}
		return res;
	}
	
	/**
	 * Encrypts a list of values using a list of keys, this keys are hashed based on their location
	 * @param list list of items to encrypts
	 * @param keys hashed keys
	 * @return A list of encrypted items
	 */
	
	public final static LinkedList<BigInteger> 
	calculateY(final LinkedList<BigInteger> list, final LinkedList<BigInteger[]> keys) {
		
		if(list==null) {
			System.err.println("(1.) OT_N_1, calculateY: Null input list");
			System.exit(-1);
		}
		if(keys==null) {
			System.err.println("(2.) OT_N_1, calculateY: Null keys list");
			System.exit(-1);
		}
		
		LinkedList<BigInteger> resList=new LinkedList<BigInteger>();
		int index[]=new int[keys.size()];
		BigInteger res=null;
		for(int i=0; i<list.size(); i++) {
			for(int j=0; j<index.length; j++){
				index[j]=0;
			}
			
			String bin=Integer.toBinaryString(i);
			
			for(int k=0, j=bin.length()-1; j>=0; j--,k++) {
				if(bin.charAt(j)=='1') {
					index[k]=1;
				}
			}
			
			res=F(keys.get(0)[index[0]], BigInteger.valueOf(i),FSize);
			
			for(int j=1; j<keys.size(); j++) {
				res=xorBigIntegers(res, F(keys.get(j)[index[j]], BigInteger.valueOf(i),FSize));
			}
			res=xorBigIntegers(res, list.get(i));
			resList.add(res);
		}

		return resList;
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Decrypts a value based on a list of keys.
	 * @param Y encrypted value.
	 * @param keys a list of encryption keys
	 * @param index index of Y in the list of the encrypted values.
	 * @return Decrypted value
	 */
	public final static BigInteger 
	calculateX(final BigInteger Y,
			final LinkedList<BigInteger> keys,
			int index) {
		
		if(Y==null) {
			System.err.println("(1.) OT_N_1, calculateX: Null value");
			System.exit(-1);
		}
		if(keys==null) {
			System.err.println("(2.) OT_N_1, calculateX: Null keys list");
			System.exit(-1);
		}
		
		BigInteger res=Y;
				
		for(int j=0; j<keys.size(); j++) {
			res=xorBigIntegers(res, F(keys.get(j), BigInteger.valueOf(index),FSize));
		}
		return res;
	}
	
}
