package SecureProtocol.BasicPrimitives;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.math.BigInteger;
import java.util.LinkedList;

/**
 * Client side (receiver) of OT protocol. 
 * @author Ayman Jarrous
 * @version 1.0
 */

public class OT_N_1Client {
	
	private BufferedReader br=null;
	private BufferedWriter bw=null;
	private LinkedList<BigInteger> keys=new LinkedList<BigInteger>();
	private int[] index=null;
	private int j=0;
	private int N=0;
	private int l=0;
	
	public OT_N_1Client(BufferedReader br, BufferedWriter bw) {
		if(br==null) {
			System.err.println("(1.) OT_N_1Client, constructor: Null BufferedReader");
			System.exit(-1);
		}
		if(bw==null) {
			System.err.println("(2.) OT_N_1Client, constructor: Null BufferedWriter");
			System.exit(-1);
		}
		this.br=br;
		this.bw=bw;
	}
	
	/**
	 * Receiver side.
	 * @param j index of element to learn
	 * @param size size of the list elements
	 * @return the j<sup>th</sup> element.
	 */
	public BigInteger execute(int j, int size) {
		if(size<=0) {
			System.err.println("(1.) OT_N_1Client, execute: Illegal size.");
			System.exit(-1);
		}
		if(j>=size || j<0) {
			System.err.println("(2.) OT_N_1Client, execute: The range should be between 0 and "+(size)+".");
			System.exit(-1);
		}
		l=(int)(Math.ceil(Math.log(size)/Math.log(2)));
		N=(int)Math.pow(2, l);
		
		index=new int[l];
		for(int i=0; i<l; i++) {
			index[i]=0;
		}
		String bin=Integer.toBinaryString(j);
		
		for(int k=0, i=bin.length()-1; i>=0; i--) {
			if(bin.charAt(i)=='1') { 
				index[k]=1;
			}
			k++;
		}
		
		for(int i=0; i<l; i++) {
			keys.add(OT2_1Client.execute(index[i],br,bw));
		}
		
		LinkedList<BigInteger> Y=new LinkedList<BigInteger>();
		try {
			for(int i=0; i<N; i++) {
				Y.add(new BigInteger(br.readLine()));
			}
		} catch(IOException ioe) {
			System.err.println("(3.) OT_N_1Client, execute: "+ioe.getMessage());
			System.exit(-1);
		}
		BigInteger Xi=OT_N_1.calculateX(Y.get(j), keys,j);
		System.out.println("RESULT: "+Xi);
		return Xi;
	}
	
	/**
	 * Preprocessing of the OT client side
	 * @param j index of element to learn
	 * @param size size of the list elements
	 */
	
	public void preprocessing(int j, int size) {
		if(size<=0) {
			System.err.println("(1.) OT_N_1Client, preprocessing: Size list should be possitive.");
			System.exit(-1);
		}
		if(j>=size || j<0) {
			System.err.println("(2.) OT_N_1Client, preprocessing: The range should be between 0 and "+(size)+".");
			System.exit(-1);
		}
		this.j=j;
		l=(int)Math.ceil(Math.log(size)/Math.log(2));
		N=(int)Math.pow(2, l);
		index=new int[l];
		
		String bin=Integer.toBinaryString(j);
		
		for(int k=0, i=bin.length()-1; i>=0; i--) {
			if(bin.charAt(i)=='1') { 
				index[k]=1;
			}
			k++;
		}
		
		for(int i=0; i<l; i++) {
			keys.add(OT2_1Client.execute(index[i],br,bw));
		}
	}
	
	/**
	 * Continuous phase after the preprocessing
	 * @return j<sup>th</sup> value which is learned in the preprocessing phase.
	 */
	
	public BigInteger executeAfterPreprocessing() {		
		
		LinkedList<BigInteger> Y=new LinkedList<BigInteger>();
		try {
			for(int i=0; i<N; i++) {
				Y.add(new BigInteger(br.readLine()));
			}
		} catch(IOException ioe) {
			System.err.println("(1.) OT_N_1Client, executeAfterPreprocessing: "+ioe.getMessage());
			System.exit(-1);
		}
		BigInteger Xi=OT_N_1.calculateX(Y.get(j), keys,j);
		//System.out.println("Output of OT: "+Xi);
		return Xi;
	}
}
