package SecureProtocol.BasicPrimitives;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.math.BigInteger;
import java.util.LinkedList;



public class OT_N_1Server {
	private BufferedReader br=null;
	private BufferedWriter bw=null;
	private LinkedList<BigInteger[]> keys=null;
	private LinkedList<BigInteger> hashedKeys=null;
	private int size=0;
	private int N=0;
	private int l=0;
	
	public OT_N_1Server(BufferedReader br, BufferedWriter bw) {
		if(br==null) {
			System.err.println("(1.) OT_N_1Server, constructor: Null BufferedReader");
			System.exit(-1);
		}
		if(bw==null) {
			System.err.println("(2.) OT_N_1Server, constructor: Null BufferedWriter");
			System.exit(-1);
		}
		this.br=br;
		this.bw=bw;
	}
	
	
	/**
	 * Creates l pairs of keys.
	 * @param l number of pairs of keys to create
	 * @param size size of the key.
	 * @return Linked list of keys, every element in the list is an array of size 2 
	 * (K<sup>0</sup><sub>i</sub>, K<sup>1</sup><sub>i</sub>).
	 */
	final private static LinkedList<BigInteger[]> createKeys(int l, int size) {
		if(l<0) {
			System.err.println("(1.) OT_N_1Server, createKeys: Invalid number of keys");
			System.exit(-1);
		}
		if(size<0) {
			System.err.println("(2.) OT_N_1Server, createKeys: Invalid size of keys");
			System.exit(-1);
		}
		
		BigInteger arr[]=null;
		LinkedList<BigInteger[]> list=new LinkedList<BigInteger[]>();
		for(int i=0; i<l; i++) {
			arr=new BigInteger[2];
			
			// Creating key to AES (F)
			arr[0]=new BigInteger(size, OT_N_1.RAND_GENERATOR);
			while(arr[0].bitLength()<=(size-Byte.SIZE+1)) {
				arr[0]=new BigInteger(size, OT_N_1.RAND_GENERATOR);
			}
			
			arr[1]=new BigInteger(size, OT_N_1.RAND_GENERATOR);
			while((arr[0].compareTo(arr[1])==0) || (arr[1].bitLength()<=(size-Byte.SIZE+1))) {
				arr[1]=new BigInteger(size, OT_N_1.RAND_GENERATOR);
			}
			list.add(arr);
		}
		return list;
	}

	/**
	 * Executes OT server side
	 * @param sList server list of items.
	 */
	@SuppressWarnings("unchecked")
	public void execute(final LinkedList<BigInteger> sList) {
		if(sList==null) {
			System.err.println("(1.) OT_N_1Server, execute, list is null.");
			System.exit(-1);
		}
		this.size=sList.size();
		l=(int)Math.ceil(Math.log(this.size)/Math.log(2));
		N=(int)Math.pow(2, l);
		
		LinkedList<BigInteger> list=(LinkedList<BigInteger>) sList.clone();
		// If size of the list is not exponent of 2
		if(list.size()<N) {
			for(int i=sList.size(); i<N ;i++) {
				list.add(BigInteger.ZERO);
			}
		}
		
		keys= createKeys(l, OT_N_1.AESSise-1);
		LinkedList<BigInteger> Y=OT_N_1.calculateY(list, keys);
		
		for(int i=0; i<l; i++) {
			OT2_1Server.execute(keys.get(i)[0], keys.get(i)[1],br,bw);
		}
		try {
			for(int i=0; i<Y.size(); i++) {
				bw.write(Y.get(i).toString());
				bw.newLine();
			}
			bw.flush();
		} catch(IOException ioe) {
			System.err.println("(2.) OT_N_1, execute: "+ioe.getMessage());
			System.exit(-1);
		}
	}
	
	/**
	 * Preprocessing of OT server side.
	 * @param size number of items of the server
	 */
	public void preprocessing(int size) {
		if(size<=0) {
			System.err.println("OT_N_1Server, preprocessing, length list should be possitive.");
			System.exit(-1);
		}
		this.size=size;
		l=(int)Math.ceil(Math.log(this.size)/Math.log(2));
		N=(int)Math.pow(2, l);
		keys=createKeys(l, OT_N_1.AESSise-1);
		
		hashedKeys=OT_N_1.prepareFKey(keys);
		for(int i=0; i<l; i++) {
			OT2_1Server.execute(keys.get(i)[0], keys.get(i)[1],br,bw);
		}
	}
		
	/**
	 * Continuous phase after the preprocessing
	 * @param sList server list
	 */
	@SuppressWarnings("unchecked")
	public void executeAfterPreprocessing(final LinkedList<BigInteger> sList) {
		LinkedList<BigInteger> list=null;
		if(sList==null) {
			System.err.println("(1.) OT_N_1Server, executeAfterPreprocessing: Null list");
			System.exit(-1);
		}
		if(sList.size()>hashedKeys.size()) {
			System.err.println("(3.) OT_N_1Server, executeAfterPreprocessing: List size is too big");
			System.exit(-1);
		}
		list=(LinkedList<BigInteger>) sList.clone();
		
		if(sList.size()<N) {
			for(int i=list.size(); i<N ;i++) { 
				list.add(BigInteger.ZERO);
			}
		}		
		LinkedList<BigInteger> Y=OT_N_1.calculateYBasedonPrepareedKeys(list, hashedKeys);
		try {
			for(int i=0; i<Y.size(); i++) {
				bw.write(Y.get(i).toString());
				bw.newLine();
			}
			bw.flush();
		} 
		catch(IOException ioe) {
			System.err.println("(2.) OT_N_1Server, executeAfterPreprocessing: "+ioe.getMessage());
			System.exit(-1);
		}
	}
}
