package utils.crypto;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.math.BigInteger;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SecureRandom;


/**
 * This class implements an Homomorphic scheme. <br>
 * The implementation is based on:<br>
 * I. Damgaard, and M. Jurik: <i><a href="http://www.daimi.au.dk/~ivan/GenPaillier_finaljour.ps">
 * A Generalization, a Simplification and some Applications 
 * of Paillier's Probabilistic Public-Key System</a></i>.
 * @author Ayman Jarrous
 * @version 1.0
 */
public class HomomorphicInstance {

	private static final long serialVersionUID = 1L;

	private final HomomorphicPublicKey homomorphicPublicKey;
	
	private final BigInteger N;
	private final BigInteger N2; //N^2
	
	private final BigInteger MINUSONE;	// -1
	
	// Encrypted message
	private final BigInteger cipher;
	
	// Random generator
	private static SecureRandom RAND_GENERATOR=new SecureRandom();

    //---------------------------------------------------------------
    //---------------------------------------------------------------
	
	/**
	 * Creates a Homomorphic encryption.
	 * @param homomorphicPublicKey a Homomorphic public key to encrypt
	 * @param message a message to be encrypted.
	 */
	public HomomorphicInstance(final HomomorphicPublicKey homomorphicPublicKey,
			final BigInteger message) {
		try {
			RAND_GENERATOR=SecureRandom.getInstance("SHA1PRNG", "SUN");
		} catch (NoSuchAlgorithmException nsae) {
			System.err.println("(1.) HomomorphicInstance, constructor: "+nsae.getMessage());
			System.exit(-1);
		} catch (NoSuchProviderException nspe) {
			System.err.println("(2.) HomomorphicInstance, constructor: "+nspe.getMessage());
			System.exit(-1);
		}
		
		if(homomorphicPublicKey==null) {
			System.err.println("(3.) HomomorphicInstance, constructor: Null homomorphic public key");
			System.exit(-1);
		}
		this.homomorphicPublicKey=homomorphicPublicKey;
		
		if(message==null) {
			System.err.println("(4.) HomomorphicInstance, constructor: Null message");
			System.exit(-1);
		}
		
		N=homomorphicPublicKey.getN();
		N2=N.pow(2);
		MINUSONE=N2.subtract(BigInteger.ONE);
		
		// Find an s that is big enough 
        int s = (message.bitLength ()/N.bitLength()) + 1;
        BigInteger N_s = N.pow (s); // n^s
        
        while (N_s.compareTo (message) <= 0) {
            N_s = N_s.multiply (N);
            s++;
        }
        
        
        cipher=createHomomorphicInstance(message,s);
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Return an encryption of a message using the public key. 
     * It calculates what s should be used and calls <code>createHomomorphicInstance</code>
     * recursively with specific s. 
	 * @param message a message to be encrypted
	 * @param s the size of s used in the encryption, n<SUP>s+1</SUP> &gt; cipher or an exception is thrown.
	 * @return The encryption of the message.
	 */
	final private BigInteger createHomomorphicInstance(final BigInteger message, final int s) {
		
		BigInteger Ns=N.pow(s);
	    
		if(message==null){
			System.err.println("(1.) HomomorphicInstance, createHomomorphicInstance: Null message");
			System.exit(-1);
		}
		
        if (message.compareTo(Ns)>= 0) {
        	System.err.println("(2.) HomomorphicInstance, createHomomorphicInstance: Message too big for encryption with " + s);
        	System.exit(-1);
        }
        BigInteger Ns1=N.multiply(Ns);  // n^{s+1}
        
        
        BigInteger r=new BigInteger(Ns1.bitLength (), RAND_GENERATOR);
        while ((r.compareTo(Ns1)>= 0) ||
               (r.compareTo(BigInteger.ZERO)==0) ||
               (BigInteger.ONE.compareTo(r.gcd(N))<0)) {
        	r=new BigInteger(Ns1.bitLength(), RAND_GENERATOR);
        }
        
        BigInteger res=BigInteger.ONE.add(message.multiply(N));
        BigInteger binomial=message;
        BigInteger msgI=message;
        BigInteger bigI=BigInteger.ONE;
        BigInteger nI=N;


        
        for (int i=2; i<=s; i++) {
        	nI=nI.multiply(N);
        	msgI=msgI.subtract(BigInteger.ONE);
        	bigI=bigI.add(BigInteger.ONE);            
        	binomial=((binomial.multiply(msgI)).multiply(bigI.modInverse(Ns1))).mod(Ns1);
        	res=(res.add(binomial.multiply(nI))).mod(Ns1);
        }
        res=(res.multiply(r.modPow(Ns, Ns1))).mod(Ns1);

        return res;
	}
	
    //---------------------------------------------------------------
        
	/**
	 * Translates the representative encrypted string to a homomorphic instance.
	 * @param homomorphicPublicKey - Homomorphic public key of the encryption
	 * @param message - representative string of the encrypted message.
	 */
	public HomomorphicInstance(final HomomorphicPublicKey homomorphicPublicKey,
			final String message) {
		if(homomorphicPublicKey==null) {
			System.err.println("(1.) HomomorphicInstance, constructor: Null homomorphic private key");
			System.exit(-1);
		}
		this.homomorphicPublicKey=homomorphicPublicKey;
		
		if(message==null){
			System.err.println("(2.) HomomorphicInstance, constructor: Null message");
			System.exit(-1);
		}
		
		N=this.homomorphicPublicKey.getN();
		N2=this.homomorphicPublicKey.getN().pow(2);
		MINUSONE=N2.subtract(BigInteger.ONE);
		
		cipher=(new BigInteger(message));
	}
	
    //---------------------------------------------------------------
		
	/**
	 * Add operation between two homomorphic instance messages.
	 * @param encryptedVal - the encrypted message to add.
	 * @return The encrypted sum of the two messages. 
	 */
	public HomomorphicInstance add(final HomomorphicInstance encryptedVal) {
		
		if(encryptedVal==null) {
			System.err.println("(1.) HomomorphicInstance, add(HomomorphicInstance): Null homomorphic encrypted value");
			System.exit(-1);
		}
		if(!encryptedVal.N.equals(this.N)) {
			System.err.println("(2.) HomomorphicInstance, add(HomomorphicInstance): Not the same homomorphic public key");
			System.exit(-1);
		}
		
		return new HomomorphicInstance(this.homomorphicPublicKey,
				(this.cipher.multiply(encryptedVal.cipher)).mod(N2).toString());
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Add operation between homomorphic instance and unencrypted ({@link BigInteger}) message. 
	 * @param val - message to add.
	 * @return Encrypted sum of both messages.
	 */	
	public HomomorphicInstance add(final BigInteger val) {
		
		if(val==null) {
			System.err.println("HomomorphicInstance, add(BigInteger): Null BigInteger value");
			System.exit(-1);
		}
		BigInteger tempVal;
		if(val.compareTo(N2)>=0){
			tempVal=val.mod(N2);
		} else {
			tempVal=val;
		}
		
		return this.add(new HomomorphicInstance(this.homomorphicPublicKey, tempVal));
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Multiply operation between the encrypted message and a constant value. 
	 * @param val - constant value to multiply.
	 * @return Encrypted multiplication of both messages.
	 */
	public HomomorphicInstance multiply(final BigInteger val) {
		
		if(val==null) {
			System.err.println("HomomorphicInstance, Multiply: Null value");
			System.exit(-1);
		}
		BigInteger tempVal=null;
		if(val.compareTo(N2)>=0){
			tempVal=val.mod(N2);
		} else{
			tempVal=val;
		}
		
		return new HomomorphicInstance(this.homomorphicPublicKey,
				this.cipher.modPow(tempVal,N2).toString());
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Subtraction operation between two homomorphic instance messages.
	 * @param encryptedVal - the encrypted message to subtract.
	 * @return The encrypted subtraction of the two messages.
	 */
	public HomomorphicInstance substract(final HomomorphicInstance encryptedVal) {
	
		if(encryptedVal==null) {
			System.err.println("(1.) HomomorphicInstance, substract(HomomorphicInstance): Null homomorphic encrypted value");
			System.exit(-1);
		}
		if(!encryptedVal.N.equals(this.N)) {
			System.err.println("(2.) HomomorphicInstance, substract(HomomorphicInstance): Not the same homomorphic public key");
			System.exit(-1);
		}
		
		return new HomomorphicInstance(this.homomorphicPublicKey,
				this.cipher.multiply(encryptedVal.cipher.modPow(MINUSONE, N2)).mod(N2).toString());
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Subtraction operation between homomorphic instance and unencrypted message. 
	 * @param val - message to subtract.
	 * @return Encrypted subtraction of both messages.
	 */	
	public HomomorphicInstance substract(final BigInteger val) {
		
		if(val==null) {
			System.err.println("HomomorphicInstance, substract(BigInteger): null BigInteger value");
			System.exit(-1);
		}
		
		return this.substract(new HomomorphicInstance(this.homomorphicPublicKey, val));
	} 
	
    //---------------------------------------------------------------
	
	/**
	 * Division operation between the encrypted message and the constant value. 
	 * @param val - constant value to divide.
	 * @return encrypted division of both messages.
	 */
	public HomomorphicInstance divide(final BigInteger val) {
		
		if(val==null) {
			System.err.println("HomomorphicInstance, divide: Null value");
			System.exit(-1);
		}
			
		BigInteger valInv=val.modInverse(N2);
		return new HomomorphicInstance(this.homomorphicPublicKey,
				this.cipher.modPow(valInv,N2).toString());
	}
	
    //---------------------------------------------------------------
	
	/**
	 * @return Homomorphic public key
	 */
	public HomomorphicPublicKey getHomomorphicPublicKey() {
		
		return this.homomorphicPublicKey;
	}
	
    //---------------------------------------------------------------
	
	public String toString() {
		
		return this.cipher.toString();
	}
	
    //---------------------------------------------------------------
	
	/**
	 * @param homomorphicPublicKey - public key to encrypt.
	 * @return Encryption of zero
	 */
	public static HomomorphicInstance getZero(final HomomorphicPublicKey homomorphicPublicKey) {
		
		return new HomomorphicInstance(homomorphicPublicKey,BigInteger.ZERO);
	}
	
    //---------------------------------------------------------------
	
	/**
	 * @param homomorphicPublicKey - public key to encrypt.
	 * @return Encryption of one
	 */
	public static HomomorphicInstance getOne(final HomomorphicPublicKey homomorphicPublicKey) {
		return new HomomorphicInstance(homomorphicPublicKey,BigInteger.ONE);
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Decrypts a homomorphic instance according to a private key.
	 * @param homomorphicPrivateKey - private key to decrypt the message 
	 * @param homomorphicInstance - encrypted message
	 * @return The decryption of the homomorphic message.
	 */
	public static BigInteger 
	decrypt(final HomomorphicPrivateKey homomorphicPrivateKey,
			final HomomorphicInstance homomorphicInstance) {
		
		if(homomorphicPrivateKey==null) {
			System.err.println("(1.) HomomorphicInstance, decrypt: Null homomorphic private key");
			System.exit(-1);
		}
		
		if(homomorphicInstance==null) {
			System.err.println("(2.) HomomorphicInstance, decrypt: Null homomorphic instance");
			System.exit(-1);
		}
		
		int s =(homomorphicInstance.cipher.bitLength()-1)/(homomorphicPrivateKey.getN().bitLength());
		
        return decrypt(homomorphicPrivateKey, homomorphicInstance, s);
	}
	
    //---------------------------------------------------------------
	
	/**
	 * Function for decrypting with surrounding private key. 
     * This will be correct with 1/n probability.
     * @param homomorphicPrivateKey - private key to decrypt the message 
	 * @param homomorphicInstance - encrypted message
	 * @param s the size of s used in the decryption, n<SUP>s+1</SUP> &gt; cipher or an exception is thrown.
	 * @return The decryption of a homomorphic message
	 */
	private static BigInteger 
	decrypt(final HomomorphicPrivateKey homomorphicPrivateKey,
			final HomomorphicInstance homomorphicInstance, int s)  {
		
		BigInteger temp;
		BigInteger count;
		
		BigInteger[] n=new BigInteger[s+1];
		n[0]=homomorphicPrivateKey.getN ();
		for(int i = 0; i < s; i++)
			n[i+1]=n[i].multiply(n[0]);
		
		BigInteger lambda=homomorphicPrivateKey.getLambda();

		BigInteger d;
		d=lambda.multiply(lambda.modInverse(n[s-1]));

		BigInteger cip_d;
		BigInteger[] L=new BigInteger[s];
		BigInteger[] mult=new BigInteger[s-1]; 
		BigInteger msg;
		
		cip_d=homomorphicInstance.cipher.modPow(d, n[s]);
		
		for(int i=0; i<s; i++)
            L[i]=((cip_d.mod(n[i+1])).subtract(BigInteger.ONE)).divide(n[0]);

        temp=BigInteger.ONE;
        count=BigInteger.ONE;
        
        for(int i = 1; i < s; i++) {
        	count=count.add(BigInteger.ONE);
        	temp=temp.multiply(count);
        	mult[i-1]=(n[i-1].multiply(temp.modInverse(n[s-1]))).mod(n[s-1]);
        }

        BigInteger t1, t2; 
        msg = null;
        
        for(int j=1; j<=s; j++) {
        	t1=L[j-1];
        	t2=msg;
        	for(int k = 2; k <= j; k++) {
        		msg=msg.subtract(BigInteger.ONE);
        		t2=(t2.multiply(msg)).mod(n[j-1]);
        		t1=(t1.subtract(t2.multiply(mult[k-2].mod(n[j-1])))).mod(n[j-1]);
        	}
        	msg=t1;
        }

        return msg;
    
	}
	
    //---------------------------------------------------------------
    //---------------------------------------------------------------
    //---------------------------------------------------------------
	
	public static class HomomorphicPrivateKey {
	
		private static final long serialVersionUID = 1L;

		// p,q prime numbers
		
		public BigInteger p=new BigInteger(
				"122634309053922149405139340029514698998515917338639177290" +
				"250855899438030536963377052803418562338560425077426631759" +
				"88100070026089381761225460083668396025669",
				10);
		
		public BigInteger q=new BigInteger("123138172844951257903898" +
				"624982280170133793090118006579374193280400087919235972861" +
				"780989881726765795893304423417585138646283909606903791361" +
				"05816752478291059",
				10);
		
		// N = (p-1)(q-1)
		public final BigInteger N;
		
		// The certainty that the choosen numbers are prime
		private int certainty=64;
		
		// LCM(p,q)
		private BigInteger lambda=null;
		
		// G; order(g) is multiple of N
		private final BigInteger g;
		
		// Instance of the public homomorphic key
		private final HomomorphicPublicKey publicKey;
		
	    //---------------------------------------------------------------
	    //---------------------------------------------------------------

		/**
		 * Constructor, used for setting up the private key, it will generate two positive BigIntegers that is probably prime. 
		 * @param bitLength length of the two primes that creates the private key.
		 */
		public HomomorphicPrivateKey(final int bitLength) {
			
			if(bitLength<=1) {
				System.err.println("HomomorphicPrivateKey, constructor: Illegal length");
				System.exit(-1);
			}
			
			p=new BigInteger(bitLength/2, certainty, RAND_GENERATOR);
			
			BigInteger tempQ=null;
			while((tempQ=new BigInteger(bitLength/2, certainty, RAND_GENERATOR)).compareTo(p)==0);
			
			q=tempQ;
			
			N=p.multiply(q);
			lambda=(p.subtract(BigInteger.ONE).multiply(q.subtract(BigInteger.ONE))).divide(p.subtract(BigInteger.ONE).gcd(q.subtract(BigInteger.ONE)));//(p-1)(q-1)/gcd(p-1,q-1); lcm
			g=lambda.multiply(lambda.modInverse (N.pow(2)));//order(g) is multiple of N
			
			publicKey=new HomomorphicPublicKey(N,g);
		}
		
	    //---------------------------------------------------------------

		/**
		 * Constructor, used for setting up the private key, it will generate two positive 
		 * BigIntegers of default <i><code>l</code></i> (1024 bits), that are probably prime. 
		 */
		public HomomorphicPrivateKey() {
			N=p.multiply(q);
			lambda=(p.subtract(BigInteger.ONE).multiply(q.subtract(BigInteger.ONE))).divide(p.subtract(BigInteger.ONE).gcd(q.subtract(BigInteger.ONE)));//(p-1)(q-1)/gcd(p-1,q-1); lcm
			g=lambda.multiply (lambda.modInverse (N.pow(2)));//order(g) is multiple of N
			
			publicKey=new HomomorphicPublicKey(N,g);
		}
		
	    //---------------------------------------------------------------

		/**
		 * Constructor that gets its prime numbers from the user.
		 * @param p prime number
		 * @param q prime number
		 */
		public HomomorphicPrivateKey(BigInteger p, BigInteger q) {
			
			if(!p.isProbablePrime(certainty)) {
				System.err.println("(1.) HomomorphicPrivateKey, constructor: "+p+" is not prime number");
				System.exit(-1);
			}
			if(!q.isProbablePrime(certainty)) {
				System.err.println("(2.) HomomorphicPrivateKey, constructor: "+q+" is not prime number");
				System.exit(-1);
			}
			
			this.p=new BigInteger(p.toString());
			this.q=new BigInteger(q.toString());
			
			N=p.multiply(q);
			lambda=(p.subtract(BigInteger.ONE).multiply(q.subtract(BigInteger.ONE))).divide(p.subtract(BigInteger.ONE).gcd(q.subtract(BigInteger.ONE)));//(p-1)(q-1)
			g=lambda.multiply (lambda.modInverse (N.pow(2)));//???
			
			publicKey=new HomomorphicPublicKey(N,g);
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return The Homomorphic public key.
		 */
		public HomomorphicPublicKey getPublicKey() {
			
			return publicKey;
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return Lambda.
		 */
		public BigInteger getLambda() {
			
			return lambda;
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return N=<i>(p-1)(q-1)</i>.
		 */
		public BigInteger getN() {
			
			return N;
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return P.
		 */
		public BigInteger getP() {
			
			return p;
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return q.
		 */
		public BigInteger getQ() {
			
			return q;
		}
	}
	
    //---------------------------------------------------------------
    //---------------------------------------------------------------
    //---------------------------------------------------------------
	
	
	public static class HomomorphicPublicKey {
	
		private static final long serialVersionUID = 1L;

		// The field creator N^2
		private final BigInteger N;

		// G
		private final BigInteger g;
		
	    //---------------------------------------------------------------
	    //---------------------------------------------------------------
		
		/**
		 * Constructor, used for encapsulating the public homomorphic keys 
		 * @param N - <i>(p-1)(q-1)</i> of the private homomorphic keys
		 * @param g - belongs to [1, N<sup>2</sup>]
		 */
		HomomorphicPublicKey(final BigInteger N, final BigInteger g) throws RuntimeException {
		
			if(N==null || N.compareTo(BigInteger.ZERO)<=0) {
				System.err.println("(1.) HomomorphicPublicKey, construcor: Illegal N");
				System.exit(-1);
			}
			if(g==null || g.compareTo(BigInteger.ZERO)<=0) {
				System.err.println("(2.) HomomorphicPublicKey, construcor: Illegal g");
				System.exit(-1);
			}
			
			this.N=new BigInteger(N.toString());
			this.g=new BigInteger(g.toString());
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return N.
		 */
		public BigInteger getN() {
			
			return N;
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * @return g.
		 */
		
		public BigInteger getG() {
			
			return g;
		}
		
		
	    //---------------------------------------------------------------
	    //---------------------------------------------------------------
	    //---------------------------------------------------------------
		
		/**
		 * Writes a given homomorphic public key to to a buffer.
		 * @param bw - buffer reader
		 * @param key - homomorphic public key to write. 
		 */
		public static void writePublicKey(BufferedWriter bw, HomomorphicPublicKey key) {
			try {
				bw.write(key.N.toString());
				bw.newLine();
				bw.write(key.g.toString());
				bw.newLine();
				bw.flush();
			}
			catch(IOException ioe) {
				System.err.println("HomomorphicInstance.HomomorphicPublicKey, writePublicKey: "+ioe.getMessage());
				System.exit(-1);
			}
		}
		
	    //---------------------------------------------------------------
		
		/**
		 * Reads a homomorphic public key.
		 * @param br - buffer reader.
		 * @return A homomorphic public key.
		 */
		public static HomomorphicPublicKey readPublicKey(BufferedReader br) {
			
			try {
				BigInteger N=new BigInteger(br.readLine());
				BigInteger g=new BigInteger(br.readLine());
				return new HomomorphicPublicKey(N,g);
			}
			catch(IOException ioe) {
				System.err.print("HomomorphicInstance.HomomorphicPublicKey, readPublicKey: "+ioe.getMessage());
				System.exit(-1);
			}
			return null;
		}
	}
}